"""
Methods for interacting with influxdb.
"""


from __future__ import annotations

import influxdb

from typing import TYPE_CHECKING
from premiscale.metrics.timeseries._base import TimeSeries

if TYPE_CHECKING:
    from typing import Dict, Tuple


class InfluxDB(TimeSeries):
    """
    Implement required interface methods that connect with InfluxDB.
    """
    def __init__(self, url: str, database: str, username: str, password: str) -> None:
        self.url = url
        self.database = database
        self._connection = influxdb.InfluxDBClient(
            self.url,
            self.database,
            username,
            password
        )

    def open(self) -> None:
        """
        Open a connection to the metrics backend these methods interact with.

        Raises:
            NotImplementedError: if the method is not implemented.
        """
        raise NotImplementedError

    def close(self) -> None:
        """
        Close the connection to the metrics backend.

        Raises:
            NotImplementedError: if the method is not implemented.
        """
        raise NotImplementedError

    def get_all(self) -> Tuple:
        """
        Get all the data in the metrics store.

        Returns:
            Tuple: all the data in the metrics store.

        Raises:
            NotImplementedError: if the method is not implemented.
        """
        raise NotImplementedError

    def commit(self) -> None:
        """
        Commit any changes to the database.

        Raises:
            NotImplementedError: if the method is not implemented.
        """
        raise NotImplementedError

    def insert(self, data: Dict) -> None:
        """
        Insert a point into the metrics store.

        Args:
            data (Dict): the data to insert.

        Raises:
            NotImplementedError: if the method is not implemented.
        """
        raise NotImplementedError

    def insert_batch(self, data: Tuple) -> None:
        """
        Insert a batch of points into the metrics store.

        Args:
            data (Tuple): the data to insert.

        Raises:
            NotImplementedError: if the method is not implemented.
        """
        raise NotImplementedError

    def clear(self) -> None:
        """
        Clear the metrics store of all data.
        """
        raise NotImplementedError

    def _run_retention_policy(self) -> None:
        """
        Run the retention policy on the database, removing points older than the retention policy.

        Raises:
            NotImplementedError: if the method is not implemented.
        """
        raise NotImplementedError


"""
Reconcile metrics and state databases and place Actions on the autoscaling queue for the Autoscaling subprocess to process.
"""


from __future__ import annotations

import logging
import time

from typing import TYPE_CHECKING
from setproctitle import setproctitle
from premiscale.metrics import build_state_connection, build_timeseries_connection


if TYPE_CHECKING:
    from premiscale.config.v1alpha1 import Config
    from multiprocessing.queues import Queue


log = logging.getLogger(__name__)


class Reconcile:
    """
    Similar to metrics - a reconciliation loop that queries influxdb for the list of VMs
    metrics came from and compares these data to state stored in MySQL. If they don't match,
    actions to correct the state drift are added to the queue.
    """
    def __init__(self, config: Config) -> None:
        self.state_database = build_state_connection(config)
        self.metrics_database = build_timeseries_connection(config)
        self.platform_queue: Queue
        self.asg_queue: Queue

    def __call__(self, asg_queue: Queue, platform_queue: Queue) -> None:
        setproctitle('reconcile')
        self.asg_queue = asg_queue
        self.platform_queue = platform_queue
        log.debug('Starting reconciliation subprocess')
        while True:
            self.platform_queue.put('Hello')
            time.sleep(5)
        # Open database connections


"""
Define actions the agent can take against infrastructure.
"""


from __future__ import annotations

import enum

from typing import TYPE_CHECKING
from abc import ABC, abstractmethod
from premiscale.hypervisor._base import Libvirt


if TYPE_CHECKING:
    from typing import Any


class Verb(enum.Enum):
    """
    Classify different C[R]UD operations the controller can take on infrastructure, e.g. creating, deleting,
    or migrating VMs.
    """
    # Do nothing.
    NULL = 0

    # Creates
    CREATE = 1
    CLONE = 3

    # Updates
    MIGRATE = 2
    REPLACE = 4

    # Deletes
    DELETE = 10
    DELETE_STORAGE = 11


class Action(ABC):
    """
    Encapsulate the various actions that the autoscaler can take. These get queued up and acted upon in the
    autoscaling subprocess as threads. One action is processed at a time in each thread, and each thread
    corresponds to an ASG.
    """
    def __init__(self, action: Verb) -> None:
        self.action = action

    @abstractmethod
    def audit_trail_msg(self) -> dict:
        """
        Return a dictionary (JSON) object containing audit data about the action taken.

        Returns:
            dict: generates an audit trail message based on what occurred with the Action's execution.
        """
        return {}

    def kind(self) -> Verb:
        """
        Return the type of action.

        Returns:
            Verb: the type of action.
        """
        return self.action

    def __enter__(self) -> Action:
        return self

    def __exit__(self, *args: Any) -> None:
        pass


class Null(Action):
    """
    Action encapsulating logic to do nothing on a VM on a particular host.
    """
    def __init__(self, vm_name: str, host: str) -> None:
        super().__init__(action=Verb.NULL)


class Create(Action):
    """
    Action encapsulating logic to create a VM on a particular host.
    """
    def __init__(self, vm_name: str, host: str) -> None:
        super().__init__(action=Verb.CREATE)


class Migrate(Action):
    """
    Action encapsulating logic to migrate a VM from one host to another.
    """
    def __init__(self, vm_name: str, host: str) -> None:
        super().__init__(action=Verb.MIGRATE)


"""
Factory method for building hypervisor connections based on the user-provided configuration.
"""


from __future__ import annotations

import logging

from typing import TYPE_CHECKING
from cattr import unstructure


if TYPE_CHECKING:
    from premiscale.hypervisor._base import Libvirt
    from premiscale.config.v1alpha1 import Host


log = logging.getLogger(__name__)


def build_hypervisor_connection(host: Host, readonly: bool = False) -> Libvirt:
    """
    Build a Libvirt connection object based on the user-provided configuration of the host.
    """
    conf = unstructure(host)

    # Remove the SSH key from the configuration as it's not needed for the connection.
    # It's only actually needed for the SSH config, which is parsed and produced when the config is validated.
    del conf['sshKey']

    match host.hypervisor:
        case 'qemu':
            log.debug(f'Using QEMU hypervisor for host {host.name} at {host.address}')

            from premiscale.hypervisor.qemu import Qemu

            # Subtypes don't need to know about the hypervisor type as it's already been set.
            del conf['hypervisor']

            return Qemu(
                readonly=readonly,
                **conf
            )
        # case 'esx':
        #     log.debug(f'Using ESX hypervisor for host {host.name} at {host.address}')

        #     from premiscale.hypervisor.esx import ESX

        #     del conf['hypervisor']

        #     return ESX(
        #         readonly=readonly,
        #         **conf
        #     )
        # case 'xen':
        #     log.debug(f'Using Xen hypervisor for host {host.name} at {host.address}')

        #     from premiscale.hypervisor.xen import Xen

        #     del conf['hypervisor']

        #     return Xen(
        #         readonly=readonly,
        #         **conf
        #     )
        case _:
            raise ValueError(f'Unknown hypervisor type: {host.hypervisor}')


"""
This module is responsible for starting relevant subprocesses and main-process threads based on how the
controller is configured. It also starts the healthcheck API for Docker and Kubernetes.
"""


from __future__ import annotations

import multiprocessing as mp
import logging
import traceback

from functools import partial
from multiprocessing.queues import Queue
from concurrent.futures import ProcessPoolExecutor, as_completed
from threading import Thread
from typing import cast, TYPE_CHECKING
from setproctitle import setproctitle
from premiscale.api.healthcheck import app as healthcheck
from premiscale.autoscaling.group import Autoscaler
from premiscale.platform import Platform


if TYPE_CHECKING:
    from premiscale.config.v1alpha1 import Config


log = logging.getLogger(__name__)


def start(config: Config, version: str, token: str) -> int:
    """
    Start our subprocesses and the healthcheck API for Docker and Kubernetes.

    Args:
        config (Config): controller config file as a Config object.
        version (str): controller version.
        token (str): controller registration token.

    Returns:
        int: status code of the first subprocess to exit.
    """

    _ret_code = 0

    setproctitle('premiscale')

    # Start the healthcheck and other APIs in a separate thread off our main process as a daemon thread.
    _main_process_daemon_threads = [
        Thread(
            target=partial(
                healthcheck.run,
                host=config.controller.healthcheck.host,
                port=config.controller.healthcheck.port,
                debug=True,
                use_reloader=False
            ),
            daemon=True
        ),
    ]

    for _dthread in _main_process_daemon_threads:
        _dthread.start()

    with ProcessPoolExecutor() as executor, mp.Manager() as manager:

        autoscaling_action_queue: Queue = cast(Queue, manager.Queue())
        platform_message_queue: Queue = cast(Queue, manager.Queue())

        # Submit the core PremiScale subprocesses that don't depend on the controller mode.
        processes = [
            # Platform websocket connection subprocess. Maintains registration, connection and data stream -> premiscale platform).
            executor.submit(
                _platform,
                platform_message_queue
            ) if (_platform := Platform.register(
                version=version,
                token=token,
                host=config.controller.platform.domain,
                cacert=config.controller.platform.certificates.path
            )) is not None else None,

            # Autoscaling controller subprocess (works on Actions in the ASG queue)
            executor.submit(
                Autoscaler(config),
                autoscaling_action_queue
            )
        ]

        # Based on the mode the controller was started in (Kubernetes or standalone), we start the relevant subprocesses.
        match config.controller.mode:
            case 'kubernetes':
                from premiscale.metrics import MetricsCollector

                processes.append(
                    executor.submit(
                        MetricsCollector(
                            config,
                            # No need for time-series metrics collection in Kubernetes mode.
                            timeseries_enabled=False
                        )
                    )
                )

                from premiscale.reconciliation.kubernetes import KubernetesAutoscaler

                # Collect actions from the Kubernetes autoscaler and translate them into PremiScale Actions for
                # the Autoscaling subprocess to process.
                processes.append(
                    executor.submit(
                        KubernetesAutoscaler(config),
                        autoscaling_action_queue,
                        platform_message_queue
                    )
                )
            case 'standalone':
                from premiscale.metrics import MetricsCollector

                processes.append(
                    executor.submit(
                        MetricsCollector(
                            config,
                            # Enable time-series metrics collection in standalone mode since Reconciliation needs it.
                            timeseries_enabled=True
                        )
                    )
                )

                from premiscale.reconciliation.internal import Reconcile

                # In standalone mode, we reconcile the state of the ASG with the desired state, as determined by analyzing the
                # metrics collected by the MetricsCollector subprocess.
                processes.append(
                    executor.submit(
                        Reconcile(config),
                        autoscaling_action_queue,
                        platform_message_queue
                    )
                )
            case 'standalone-external-metrics':
                from premiscale.reconciliation.internal import Reconcile

                # In standalone mode, we reconcile the state of the ASG with the desired state, as determined by analyzing the
                # metrics collected by the MetricsCollector subprocess.
                processes.append(
                    executor.submit(
                        Reconcile(config),
                        autoscaling_action_queue,
                        platform_message_queue
                    )
                )
            case 'kubernetes-external-metrics':
                from premiscale.reconciliation.kubernetes import KubernetesAutoscaler

                # Collect actions from the Kubernetes autoscaler and translate them into PremiScale Actions for
                # the Autoscaling subprocess to process.
                processes.append(
                    executor.submit(
                        KubernetesAutoscaler(config),
                        autoscaling_action_queue,
                        platform_message_queue
                    )
                )

        filtered_processes = [process for process in processes if process is not None]

        for process in as_completed(filtered_processes):
            try:
                process.result()
            except Exception:
                log.error(f'Process {process} failed. Full traceback: {traceback.format_exc()}')

                for _process in filtered_processes:
                    if _process is not process:
                        _process.cancel()

                _ret_code = 1

                break

    for thread in _main_process_daemon_threads:
        thread.join(timeout=5)

    return _ret_code


"""
Methods relating to connecting to PremiScale's platform.
"""


from __future__ import annotations

import asyncio
import logging
import time
import ssl
import requests
import json
import ssl

from typing import Dict
from websockets import client as ws, exceptions as wse
from socket import gaierror
from multiprocessing.queues import Queue
from urllib.parse import urljoin
from http import HTTPStatus
from setproctitle import setproctitle
from premiscale.exceptions import RateLimitedError
from premiscale.utils import write_json, read_json
from premiscale.platform.utils import retry


log = logging.getLogger(__name__)


class Platform:
    """
    Handle communication to and from the platform. Maintains an async websocket
    connection and calls setters and getters on the other daemon threads' objects to
    configure them.

    Args:
        registration (Dict[str, str]): registration response from the registration service.
        version (str): controller version.
        host (str): platform host.
        wspath (str): path to the websocket endpoint.
        cacert (str): path to the certificate file (for use with self-signed certificates).
    """
    def __init__(self,
            registration: Dict[str, str],
            version: str,
            host: str = 'app.premiscale.com',
            wspath: str = '/agent/websocket',
            cacert: str = ''
        ) -> None:

        self.host = urljoin(
            f'wss://{host}',
            wspath
        )

        self.version = version
        self._registration = registration
        self._cacert = cacert

        self._queue: Queue
        self._received_platform_messages: asyncio.Queue = asyncio.Queue()
        self._websocket: ws.WebSocketClientProtocol

    @classmethod
    @retry(retries=3, retry_delay=5.0)
    def register(
            cls,
            token: str,
            version: str,
            host: str,
            registration_path: str = '/agent/registration',
            websocket_path: str = '/agent/websocket',
            cacert: str = ''
        ) -> Platform | None:
        """
        Register the controller with the PremiScale platform before starting this process. If registration fails, this
        classmethod returns None to the upstream caller.

        Args:
            token (str): registration API token.
            version (str): agent version.
            host (str): platform domain.
            registration_path (str): path to the registration endpoint.
            websocket_path (str): path to the controller websocket endpoint.
            cacert (str): path to the CA certificate file, if provided.

        Returns:
            Platform | None: registration service response, or an empty dict if the registration was not successful.

        Raises:
            RateLimitedError: if the request is rate limited.
        """

        # If the controller has already registered, skip the registration request.
        if (cached_registration_response := read_json('registration.json')) is not None and cached_registration_response.get('host') == host:
            log.info(f'Agent already registered with PremiScale platform at "{host}". Skipping controller registration request')
        else:
            # If no token is provided and a cached registration response is not found, simply start the controller
            # in standalone mode for development purposes.
            if token == '':
                log.warning('No registration token provided, starting controller in standalone mode')
                return None

            try:
                response = requests.post(
                    url=urljoin(
                        f'https://{host}',
                        registration_path
                    ),

                    data=json.dumps(
                        {
                            'version': version,
                            'type': 'agent',
                            'registration_key': token
                        }
                    ),
                    headers={
                        # 'Authorization': header,
                        'Content-Type': 'application/json'
                    },
                    verify=cacert
                )

                if response.status_code != HTTPStatus.OK:
                    log.error(f'Failed to register with PremiScale platform: {response.text}')
                    return None

                registration_response = response.json()

                log.debug(f'Registration response: {json.dumps(registration_response)}')

                # Append the host used to make the registration request to the registration response.
                registration_response['host'] = host

                # Write the registration response to disk for future reference; if the controller is restarted, it will
                # Attempt to read this file to determine if it has already registered. If the host changes, the controller
                # will re-register and be assigned a new agent ID by the platform.
                write_json(registration_response, '$HOME/registration.json')

            except (ssl.SSLCertVerificationError, requests.exceptions.SSLError) as msg:
                log.error(f'Could not verify SSL certificate: {msg}. Skipping registration')

                return None
            except json.JSONDecodeError:
                if response.status_code == HTTPStatus.TOO_MANY_REQUESTS:
                    # Kick a RateLimitedError exception to the retry decorator to attempt the request again.
                    raise RateLimitedError(delay=float(response.headers['x-rate-limit-reset']) / 1_000)
                else:
                    log.error(
                        f'Did not get valid JSON in response: {response.text if response.text else response.reason} ~ {response.status_code}'
                    )

                    return None

        return cls(
            registration=registration_response,
            version=version,
            host=host,
            wspath=websocket_path,
            cacert=cacert
        )

    def __call__(self, platform_queue: Queue) -> None:
        setproctitle('platform')
        self._queue = platform_queue
        log.debug('Starting platform connection subprocess')
        # This should never exit. Process should stay open forever.
        asyncio.run(
            self._set_up_connection()
        )

    async def _set_up_connection(self) -> None:
        """
        Establish websocket connection to PremiScale's platform.
        """
        ssl_context: ssl.SSLContext | None = None

        # Handle self-signed certificates if provided. Build a custom SSL context.
        if self._cacert != '':
            ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
            ssl_context.load_verify_locations(
                cafile=self._cacert
            )

        while True:
            try:
                async with ws.connect(self.host, ping_timeout=300, ssl=ssl_context) as self._websocket:
                    log.info(f'Established connection to platform hosted at \'{self.host}\'')

                    try:
                        await asyncio.gather(
                            self._recv_message(),
                            self._sync_platform_queue()
                        )
                    except wse.ConnectionClosed:
                        log.error(f'Websocket connection to \'{self.host}\' closed unexpectedly, reconnecting...')
            except (gaierror, ssl.SSLError) as msg:
                log.error(f'Could not connect to \'{self.host}\', retrying: {msg}')
                time.sleep(2)

    async def _sync_platform_queue(self) -> None:
        """
        Sync the platform queue with the platform. If this function returns, then the queue is empty.
        """
        while True:
            if not self._queue.empty():
                for _ in range(self._queue.qsize()):
                    msg = self._queue.get()
                    await self.send_message(msg)
            await asyncio.sleep(1)

    async def _recv_message(self) -> None:
        """
        Receive messages from the platform and place them on a separate queue to be acted upon.
        """
        async for msg in self._websocket:
            await self._received_platform_messages.put(msg)

    async def sync_actions(self) -> bool:
        """
        Sync actions taken by the controller for auditing.

        Returns:
            bool: True if the sync was successful.
        """
        return False

    async def sync_metrics(self) -> bool:
        """
        Sync metrics to the platform.

        Returns:
            bool: True if the sync was successful.
        """
        return False

    async def send_message(self, msg: str) -> bool:
        """
        Asynchronously send a message up to the platform.

        Args:
            msg (str): Message to send.

        Returns:
            bool: True if the item was queued.
        """
        self._queue.put(msg)
        return True


"""
Utility methods for the platform module.
"""


import time
import logging

from typing import Callable, Any, Dict
from functools import wraps
from urllib.error import URLError
from premiscale.exceptions import RateLimitedError


log = logging.getLogger(__name__)


def retry(retries: int =0, retry_delay: float = 1.0, ratelimit_buffer: float = 0.25) -> Callable:
    """
    A request retry decorator that catches common request exceptions and retries the wrapped function call.

    Args:
        retries (int): number of times to retry the wrapped function call. When `0`, retries indefinitely. (default: 0)
        retry_delay (float): if retries is 0, this delay value (in seconds) is used between retries. (default: 1.0)
        ratelimit_buffer (float): a buffer (in seconds) to add to the delay when a rate limit is hit. (default: 0.25)

    Returns:
        Callable: a decorator for the wrapped function.

    Raises:
        ValueError: if `retries` is less than 0.
    """
    if retries < 0:
        raise ValueError(f'Expected positive `retries` values, received: "{retries}"')

    def decorator(f: Callable) -> Callable:
        @wraps(f)
        def wrapper(*args: Any, **kwargs: Any) -> Dict | None:
            res: Any = None

            def call() -> Dict[str, str] | None:
                nonlocal res

                try:
                    return f(*args, **kwargs)
                except RateLimitedError as msg:
                    log.warning(f'Ratelimited, waiting "{msg.delay + ratelimit_buffer}"s before trying again')
                    time.sleep(msg.delay + ratelimit_buffer)
                    return None
                except URLError as msg:
                    log.warning(msg)
                    return None

            if retries > 0:
                # Finite number of user-specified retries.
                for _ in range(retries):
                    if (res := call()) is not None:
                        return res
                else:
                    log.error(f'Retry attempt limit exceeded')
                    return None
            else:
                # Infinite retries.
                while (res := call()) is None:
                    time.sleep(retry_delay)

                return res

        return wrapper
    return decorator


"""
Implement a Libvirt connection to an ESX-based hypervisor/host.
"""


from __future__ import annotations

from typing import TYPE_CHECKING
from premiscale.hypervisor._base import Libvirt
from premiscale.hypervisor._schemas import DomainStats


if TYPE_CHECKING:
    from typing import Dict, List
    from ipaddress import IPv4Address


# class ESX(Libvirt):
#     """
#     Connect to an ESX hypervisor.
#     """

#     def __init__(self,
#                  name: str,
#                  address: IPv4Address,
#                  port: int,
#                  protocol: str,
#                  timeout: int = 60,
#                  user: str | None = None,
#                  readonly: bool = False,
#                  resources: Dict | None = None) -> None:
#         super().__init__(
#             name=name,
#             address=address,
#             port=port,
#             protocol=protocol,
#             hypervisor='esx',
#             timeout=timeout,
#             user=user,
#             readonly=readonly,
#             resources=resources
#         )

#     def _getHostState(self) -> Dict:
#         """
#         Get the state of the VMs on the host.

#         Returns:
#             Dict: The state of the VMs on the host.
#         """
#         return {}

#     def _getHostStats(self) -> Dict:
#         """
#         Get a report of schedulable resource utilization on the host.

#         Returns:
#             Dict: The resources available on the host.
#         """
#         return {}

#     def _getHostVMStats(self) -> List[DomainStats]:
#         """
#         Get a report of resource utilization for a VM.

#         Returns:
#             List[DomainStats]: Stats of all VMs on this particular host connection.
#         """
#         return []

#     def statsToMetricsDB(self) -> None:
#         return None

#     def statsToStateDB(self) -> None:
#         return None


"""
Make the environment variables' values that we care about, portable.
"""


from __future__ import annotations

from typing import TYPE_CHECKING
from importlib import metadata as meta

import os
import logging


if TYPE_CHECKING:
    from typing import Dict


log = logging.getLogger(__name__)


env: Dict[str, str] = {
    # Environment variables with default values to configure PremiScale.
    'PREMISCALE_TOKEN': os.getenv('PREMISCALE_TOKEN', ''),
    'PREMISCALE_CONFIG_PATH': os.getenv('PREMISCALE_CONFIG_PATH', '/opt/premiscale/config.yaml'),
    'PREMISCALE_CACERT': os.getenv('PREMISCALE_CACERT', ''),
    'PREMISCALE_LOG_LEVEL': os.getenv('PREMISCALE_LOG_LEVEL', 'info'),
    'PREMISCALE_LOG_FILE': os.getenv('PREMISCALE_LOG_FILE', '/opt/premiscale/controller.log'),
}

version: str = meta.version('premiscale')


"""
Serve a /metrics-endpoint for Prometheus metrics if enabled.
"""


from __future__ import annotations

from typing import TYPE_CHECKING
from werkzeug.middleware.dispatcher import DispatcherMiddleware
from prometheus_client import make_wsgi_app


if TYPE_CHECKING:
    from flask import Flask


def create_metrics_api(app: Flask) -> None:
    """
    Create a metrics API endpoint monitoring the health of the controller.

    Args:
        app (Flask): Flask app instance.
    """

    # app.wsgi_app = DispatcherMiddleware(
    #     app.wsgi_app,
    #     {
    #         '/metrics': make_wsgi_app()
    #     }
    # )


"""
This module is responsible for the healthcheck API.
"""


from flask import Flask
from flask_cors import CORS


app = Flask(__name__)
CORS(app)


__all__ = ['app']


from premiscale.api.healthcheck.routes import create_healthcheck_api


create_healthcheck_api(app)


"""
The PremiScale autoscaling controller for Kubernetes.
"""


from __future__ import annotations

import sys
import logging

from pathlib import Path
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from premiscale.config.parse import validateConfig, configParse
from premiscale.daemon import start
from premiscale.utils import LogLevel
from premiscale import env, version


log = logging.getLogger(__name__)


def main() -> None:
    """
    Set up the CLI for PremiScale autoscaler.
    """
    parser = ArgumentParser(
        formatter_class=RawDescriptionHelpFormatter,
        description=__doc__,
        epilog='For more information, visit https://premiscale.com.\n\n© PremiScale, Inc. 2024.'
    )

    parser.add_argument(
        '--token', type=str, default=env['PREMISCALE_TOKEN'],
        help='Platform registration token. Also available as the environment variable \'PREMISCALE_TOKEN\'. If no token is provided, the controller will not register with the platform and start in standalone mode.'
    )

    parser.add_argument(
        '-c', '--config', type=str, default=env['PREMISCALE_CONFIG_PATH'],
        help='Configuration file path to use. Also available as the environment variable \'PREMISCALE_CONFIG_PATH\'. (default: /opt/premiscale/config.yaml)'
    )

    parser.add_argument(
        '--validate', action='store_true',
        help='Validate the provided configuration file and exit. (default: false)'
    )

    parser.add_argument(
        '--version', action='store_true',
        help='Display controller version.'
    )

    parser.add_argument(
        '--log-level', default=env['PREMISCALE_LOG_LEVEL'], choices=list(LogLevel), type=LogLevel.from_string,
        help='Set the logging level. Also available as the environment variable \'PREMISCALE_LOG_LEVEL\'. (default: info)'
    )

    log_group = parser.add_mutually_exclusive_group()

    log_group.add_argument(
        '--log-file', type=str, default='/opt/premiscale/controller.log',
        help='Specify the file the service logs to if --log-stdout is not set. Also available as the environment variable \'PREMISCALE_LOG_FILE\'. (default: /opt/premiscale/controller.log)'
    )

    log_group.add_argument(
        '--log-stdout', action='store_true',
        help='Log to stdout (for use in containerized deployments). (default: false)'
    )

    args = parser.parse_args()

    if args.version:
        print(f'premiscale v{version}')
        sys.exit(0)

    # Configure logger
    if args.log_stdout:
        logging.basicConfig(
            stream=sys.stdout,
            format='%(asctime)s | %(levelname)s | %(name)s | %(message)s',
            level=args.log_level.value
        )
    else:
        try:
            # Instantiate log path (when logging locally).
            if not Path(args.log_file).exists():
                Path(args.log_file).parent.mkdir(parents=True, exist_ok=True)

            logging.basicConfig(
                filename=args.log_file,
                format='%(asctime)s | %(levelname)s | %(name)s | %(message)s',
                level=args.log_level.value,
                filemode='a'
            )
        except (FileNotFoundError, PermissionError) as msg:
            log.error(f'Failed to configure logging, received: {msg}')
            sys.exit(1)

    if args.validate:
        sys.exit(0 if validateConfig(args.config) else 1)

    log.info(f'Starting PremiScale controller v{version}')

    if (token := args.token) != '':
        log.info('Using provided platform token for registration')
    else:
        token = ''

    # Start the premiscale controller.
    sys.exit(
        start(
            config=configParse(args.config),
            version=version,
            token=token
        )
    )


"""
Implement a Libvirt connection to a Qemu-based hypervisor/host.
"""


from __future__ import annotations

import logging
import re

from typing import TYPE_CHECKING, List, Tuple
from libvirt import (
    libvirtError,
    VIR_DOMAIN_NOSTATE,      # 0
    VIR_DOMAIN_RUNNING,      # 1
)
from xmltodict import parse as xmlparse
from cachetools import cached, TTLCache
from cattrs import structure
from premiscale.hypervisor._base import Libvirt
from premiscale.hypervisor.qemu_data import DomainStats

if TYPE_CHECKING:
    from typing import Dict
    from ipaddress import IPv4Address


log = logging.getLogger(__name__)


class Qemu(Libvirt):
    """
    A subclass for interacting with a Qemu-based hypervisor/host.
    """

    def __init__(self,
                 name: str,
                 address: IPv4Address,
                 port: int,
                 protocol: str,
                 timeout: int = 60,
                 user: str | None = None,
                 readonly: bool = False,
                 resources: Dict | None = None) -> None:

        super().__init__(
            name=name,
            address=address,
            port=port,
            protocol=protocol,
            hypervisor='qemu',
            timeout=timeout,
            user=user,
            readonly=readonly,
            resources=resources
        )

    @cached(cache=TTLCache(maxsize=1, ttl=5))
    def _getHostStats(self) -> Dict:
        """
        Get a report of schedulable resource utilization on the host.

        Returns:
            Dict: The resources available on the host.
        """
        if self._connection is None:
            return {}

        _stats = {
            'host': {
                'name': self._connection.getHostname(),
                'type': self._connection.getType(),
                'uri': self._connection.getURI(),
                'version': self._connection.getVersion(),
                'libvirt_version': self._connection.getLibVersion(),
                'capabilities': xmlparse(self._connection.getCapabilities()),
                'node_info': self._connection.getInfo(),
                'max_vcpus': self._connection.getMaxVcpus(None),
                'free_memory': self._connection.getFreeMemory(),
                'node_memory': self._connection.getMemoryStats(-1, 0),
                'node_cpu_stats': self._connection.getCPUStats(True),
                'stats': {
                    'cpu': self._connection.getCPUStats(
                        cpuNum=-1,
                        flags=VIR_DOMAIN_NOSTATE
                    ),
                    'memory': self._connection.getMemoryStats(
                        cellNum=-1,
                        flags=VIR_DOMAIN_NOSTATE
                    )
                }
            },
            # Get a picture of the state of the host.
            'vms': [
                {
                    'name': vm.name(),
                    'state': vm.info()
                } for vm in self._connection.listAllDomains(flags=VIR_DOMAIN_NOSTATE)
            ]
        }

        return _stats

    @cached(cache=TTLCache(maxsize=1, ttl=5))
    def _getHostVMStats(self) -> List[DomainStats]:
        """
        Get a report of resource utilization for a VM. A typical report includes all the following fields ~

            https://github.com/premiscale/premiscale/pull/196#issuecomment-2168388982

        And these fields are parsed into a DomainStats object.

        Returns:
            List[DomainStats]: Stats of all VMs on this particular host connection.
        """

        MIDDLE_NUMBER_REGEX = re.compile(r'^(vcpu|block|net)_[0-9]+_')

        if self._connection is None:
            return []

        _all_domain_stats = self._connection.getAllDomainStats(
            # https://vscode.dev/github/premiscale/premiscale/blob/store-vm-dataremiscale/lib/python3.10/site-packages/libvirt.py#L6424-L6425
            # Using 0 for @stats returns all stats groups supported by the given hypervisor. See the link
            # above for an example of all the retrieved stats.
            stats=0,
            flags=VIR_DOMAIN_RUNNING
        )

        # Normalize domains' stat fields into a list of DomainStats objects.
        domain_stats_filtered_list: List[DomainStats] = []

        for (domain, stat) in _all_domain_stats:
            stat['name'] = domain.name()

            domain_keys = list(stat.keys())

            # Set up a bunch of empty iterables to store bins of parsed stats from this domain.
            # These will be unpacked into the DomainStats object.
            domain_stats_filtered = {}
            vcpus: List[Dict[str, int]] = []
            blocks: List[Dict[str, int]] = []
            nets: List[Dict[str, int]] = []

            # Normalize all the returned stats fields and parse them into a DomainStats object.
            for key in domain_keys:
                # Replace '-' and '.' with '_' in the keys for consistency.
                oldValue = stat.pop(key)
                key = key.replace('-', '_')
                key = key.replace('.', '_')
                stat[key] = oldValue

                # Bin parsed vCPU, Block, and Net stats to form the proper hierarchy.
                if key.startswith('vcpu_') and key != 'vcpu_current' and key != 'vcpu_maximum':
                    try:
                        index = int(key.split('_')[1])
                    except IndexError as e:
                        log.error(f'IndexError: {e}')
                        return []

                    vcpus.extend([{}] * (index - len(vcpus) + 1))

                    # Remove the number from the key and strip 'vcpu_' from the beginning.
                    _new_key = re.sub(MIDDLE_NUMBER_REGEX, '', key)

                    vcpus[index][_new_key] = stat[key]
                #
                elif key.startswith('block_') and key != 'block_count':
                    try:
                        index = int(key.split('_')[1])
                    except IndexError as e:
                        log.error(f'IndexError: {e}')
                        return []

                    blocks.extend([{}] * (index - len(blocks) + 1))

                    _new_key = re.sub(MIDDLE_NUMBER_REGEX, '', key)

                    blocks[index][_new_key] = stat[key]
                #
                elif key.startswith('net_') and key != 'net_count':
                    try:
                        index = int(key.split('_')[1])
                    except IndexError as e:
                        log.error(f'IndexError: {e}')
                        return []

                    nets.extend([{}] * (index - len(nets) + 1))

                    _new_key = re.sub(MIDDLE_NUMBER_REGEX, '', key)

                    nets[index][_new_key] = stat[key]
                else:
                    domain_stats_filtered[key] = stat[key]

            domain_stats_filtered['vcpu'] = vcpus
            domain_stats_filtered['block'] = blocks
            domain_stats_filtered['net'] = nets
            domain_stats_filtered['host'] = self.name
            domain_stats_filtered['address'] = self.address

            domain_stats_filtered_list.append(
                structure(domain_stats_filtered, DomainStats)
            )

        return domain_stats_filtered_list

    def statsToStateDB(self) -> List[Tuple]:
        """
        Convert the stats from the host into a state database entry. Instead of relying on the calling class to
        format these correctly, every interface is required to implement its own method to do so, since it's not
        guaranteed that the stats will be the same across different hypervisors.

        Returns:
            List[Tuple]: The state of the host and VMs on it.
        """
        return []

    def statsToMetricsDB(self) -> List[Tuple]:
        """
        Convert the stats from the host into a metrics database entry. Instead of relying on the calling class to
        format these correctly, every interface is required to implement its own method to do so, since it's not
        guaranteed that the stats will be the same across different hypervisors.

        Returns:
            List[Tuple]: Stats to a list of metrics database entries.
        """

        vm_stats: List[DomainStats] = self._getHostVMStats()

        log.debug(f'VM Stats: {vm_stats}')

        # TODO: Implement a method to convert the host stats into a metrics database entry.
        host_stats: Dict = self._getHostStats()

        tinflux_vm_stats = [vm.to_tinyflux() for vm in vm_stats]

        return tinflux_vm_stats


"""
ABCs for metrics storage methods.
"""


from __future__ import annotations

import logging

from typing import TYPE_CHECKING
from abc import ABC, abstractmethod

if TYPE_CHECKING:
    from typing import Any, Dict, Tuple


log = logging.getLogger(__name__)


class TimeSeries(ABC):
    """
    An abstract base class with a skeleton interface for metrics class-types.
    """

    @abstractmethod
    def open(self) -> None:
        """
        Open a connection to the metrics backend these methods interact with.

        Raises:
            NotImplementedError: if the method is not implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def close(self) -> None:
        """
        Close the connection to the metrics backend.

        Raises:
            NotImplementedError: if the method is not implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def commit(self) -> None:
        """
        Commit any changes to the database.

        Raises:
            NotImplementedError: if the method is not implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def insert(self, data: Dict) -> None:
        """
        Insert a point into the metrics store.

        Args:
            data (Dict): the data to insert.

        Raises:
            NotImplementedError: if the method is not implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def insert_batch(self, data: Tuple[Dict]) -> None:
        """
        Insert a batch of points into the metrics store.

        Args:
            data (Tuple[Dict]): the data to insert.

        Raises:
            NotImplementedError: if the method is not implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def clear(self) -> None:
        """
        Clear the metrics store of all data.

        Raises:
            NotImplementedError: if the method is not implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def _run_retention_policy(self) -> None:
        """
        Run the retention policy on the database, removing points older than the retention policy.

        Raises:
            NotImplementedError: if the method is not implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def get_all(self) -> Tuple:
        """
        Get all the data in the metrics store.

        Returns:
            Tuple: all the data in the metrics store.

        Raises:
            NotImplementedError: if the method is not implemented.
        """
        raise NotImplementedError

    def __enter__(self) -> TimeSeries:
        self.open()
        return self

    def __exit__(self, *args: Any) -> None:
        self.close()
        return


"""
Process queues of Actions for all autoscaling groups. This is the main entry point for the autoscaling subprocess.
"""


from __future__ import annotations

import logging

from typing import TYPE_CHECKING
from multiprocessing.queues import Queue
from setproctitle import setproctitle
from premiscale.autoscaling.actions import Action


if TYPE_CHECKING:
    from premiscale.config.v1alpha1 import Config


log = logging.getLogger(__name__)


class Autoscaler:
    """
    Handle actions. E.g., if a new VM needs to be created or deleted on some host,
    handle that action, and all relevant side-effects (e.g. updating MySQL state).

    One of these classes gets instantiated for every autoscaling group defined in
    the config.
    """
    def __init__(self, config: Config) -> None:
        self.config = config
        self.queue: Queue

    def __call__(self, asg_queue: Queue) -> None:
        setproctitle('autoscaling')
        self.queue = asg_queue
        log.debug('Starting autoscaling subprocess')


"""
Provide methods to interact with the Libvirt API.
"""


from __future__ import annotations

import libvirt as lv
import logging

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING
from libvirt import libvirtError
from ipaddress import IPv4Address


if TYPE_CHECKING:
    from premiscale.hypervisor.qemu_data import DomainStats
    from typing import Any, Dict, List, Tuple


log = logging.getLogger(__name__)


class Libvirt(ABC):
    """
    Connect to hosts and provide an interface for interacting with VMs on them.

    Args:
        name (str): Name of the host.
        address (IPv4Address): IP address of the host to connect to.
        port (int): Port to connect to the host on.
        protocol (str): Type of authentication to use. Defaults to 'ssh'. Can be either 'ssh' or 'tls'.
        hypervisor (str): Type of hypervisor to connect to.
        timeout (int): Timeout for the connection. Defaults to 30 seconds.
        user (str | None): Username to authenticate with (if using SSH).
        readonly (bool): Whether to open the connection in read-only mode. Defaults to False.
        resources (Dict | None): Resources available on the host. Defaults to None.
    """
    def __init__(self,
                 name: str,
                 address: IPv4Address,
                 port: int,
                 protocol: str,
                 hypervisor: str,
                 timeout: int = 30,
                 user: str | None = None,
                 readonly: bool = False,
                 resources: Dict | None = None) -> None:
        self.name = name
        self.address = address
        self._address_str = str(address)
        self.port = port
        self.protocol = protocol
        self.timeout = timeout
        self.hypervisor = hypervisor

        # Set the user to 'root' if not provided by the end user.
        if user is None:
            self.user = 'root'
        else:
            self.user = user

        self.readonly = readonly
        self.resources = resources
        self._connection: lv.virConnect | None = None
        if protocol.lower() == 'ssh':
            # SSH
            self.connection_string = f'{hypervisor}+ssh://{user}@{address}:{port}/system'
        else:
            # TLS
            self.connection_string = f'{hypervisor}+tls://{address}:{port}/system'

    def __enter__(self) -> Libvirt | None:
        return self.open()

    def __exit__(self, *args: Any) -> None:
        self.close()

    def open(self) -> Libvirt | None:
        """
        Open a connection to the Libvirt hypervisor.
        """
        try:
            log.debug(f'Attempting to connect to host at {self.connection_string}')

            if self.readonly:
                self._connection = lv.openReadOnly(self.connection_string)
            else:
                self._connection = lv.open(self.connection_string)
                log.info(f'Connected to host at {self.connection_string}')
        except libvirtError as e:
            log.error(f'Failed to connect to host at {self.connection_string}: {e}')
            return None

        return self

    def close(self) -> None:
        """
        Close the connection with the Libvirt hypervisor.
        """
        if self._connection:
            self._connection.close()
            log.debug(f'Closed connection to host at {self.connection_string}')
        else:
            log.error(f'No host connection to close, probably due to an error on connection open')

    @abstractmethod
    def _getHostStats(self) -> Dict:
        """
        Get a report of schedulable resource utilization on the host.

        Returns:
            Dict: The resources available on the host.

        Raises:
            NotImplementedError: If the method is not implemented by the subclass.
        """
        raise NotImplementedError

    @abstractmethod
    def _getHostVMStats(self) -> List[DomainStats]:
        """
        Get a report of resource utilization for a VM.

        Returns:
            List[DomainStats]: Stats of all VMs on this particular host connection.

        Raises:
            NotImplementedError: If the method is not implemented by the subclass.
        """
        raise NotImplementedError

    @abstractmethod
    def statsToStateDB(self) -> List[Tuple]:
        """
        Convert the stats from the host into a state database entry. Instead of relying on the calling class to
        format these correctly, every interface is required to implement its own method to do so, since it's not
        guaranteed that the stats will be the same across different hypervisors.

        Returns:
            List[Tuple]: The state of the host and VMs on it.

        Raises:
            NotImplementedError: If the method is not implemented by the subclass.
        """
        raise NotImplementedError

    @abstractmethod
    def statsToMetricsDB(self) -> List[Tuple]:
        """
        Convert the stats from the host into a metrics database entry. Instead of relying on the calling class to
        format these correctly, every interface is required to implement its own method to do so, since it's not
        guaranteed that the stats will be the same across different hypervisors.

        Returns:
            List[Tuple]: The metrics for the host and VMs on it.

        Raises:
            NotImplementedError: If the method is not implemented by the subclass.
        """
        raise NotImplementedError


"""
Parse v1alpha1 configuration files into a Config object with attrs and cattrs.
"""


from __future__ import annotations

import logging
import os
import sys

from attrs import define
from attr import ib
from cattrs import structure

# In this particular module, cattrs requires these types during runtime to unpack,
# so we skip the TYPE_CHECKING check wrapping these imports.
from typing import Dict, List


log = logging.getLogger(__name__)


@define
class DatabaseCredentials:
    """
    Database credentials.
    """
    username: str
    password: str

    def __attrs_post_init__(self):
        """
        Post-initialization method to expand environment variables.
        """
        self.expand()

    def expand(self):
        """
        Expand environment variables in the database credentials.
        """
        self.username = os.path.expandvars(self.username)
        self.password = os.path.expandvars(self.password)


@define
class Connection:
    """
    Connection configuration options.
    """
    url: str
    database: str
    credentials: DatabaseCredentials

    def __attrs_post_init__(self):
        """
        Post-initialization method to expand environment variables.
        """
        self.expand()

    def expand(self):
        """
        Expand environment variables in the connection configuration.
        """
        self.url = os.path.expandvars(self.url)
        self.database = os.path.expandvars(self.database)


@define
class State:
    """
    State database configuration options.
    """
    type: str
    connection: Connection | None = ib(default=None)


@define
class TimeSeries:
    """
    Time series database configuration options.
    """
    type: str
    trailing: int
    dbfile: str | None = ib(default=None)
    connection: Connection | None = ib(default=None)


@define
class Databases:
    """
    Databases configuration options.
    """
    collectionInterval: int
    hostConnectionTimeout: int
    maxHostConnectionThreads: int
    state: State
    timeseries: TimeSeries
    hostConnectionQueueSize: int | None = ib(default=None)

    def __attrs_post_init__(self):
        """
        Post-initialization method to expand environment variables.
        """
        if self.hostConnectionQueueSize is None:
            self.hostConnectionQueueSize = self.maxHostConnectionThreads
        elif self.hostConnectionQueueSize < self.maxHostConnectionThreads:
            log.warning(f'Host connection queue size must be greater than or equal to the maximum host connection threads. Defaulting to {self.maxHostConnectionThreads}.')
            self.hostConnectionQueueSize = self.maxHostConnectionThreads


@define
class Certificates:
    """
    Certificate configuration options.
    """
    path: str

    def __attrs_post_init__(self):
        """
        Post-initialization method to expand environment variables.
        """
        self.expand()

    def expand(self):
        """
        Expand environment variables in the certificate configuration.
        """
        self.path = os.path.expandvars(self.path)


@define
class Platform:
    """
    Platform configuration options.
    """
    domain: str
    token: str
    certificates: Certificates
    actionsQueueMaxSize: int

    def __attrs_post_init__(self):
        """
        Post-initialization method to expand environment variables.
        """
        self.expand()

    def expand(self):
        """
        Expand environment variables in the platform configuration.
        """
        self.domain = os.path.expandvars(self.domain)
        self.token = os.path.expandvars(self.token)


@define
class Reconciliation:
    """
    Reconciliation configuration options.
    """
    interval: int


@define
class Resources:
    """
    Resource configuration options.
    """
    cpu: int
    memory: int
    storage: int


@define
class Host:
    """
    Host configuration options.
    """
    name: str
    address: str
    protocol: str
    port: int
    hypervisor: str
    sshKey: str | None = ib(default=None)  # Expected to contain the actual key or environment variable with the contents of the private key.
    timeout: int = ib(default=45)
    user: str | None = ib(default=None)
    resources: Resources | None = ib(default=None)

    def __attrs_post_init__(self):
        """
        Post-initialization method to expand environment variables.
        """
        self.expand()

        # Make sure that this call doesn't rely on any values that are updated past this point.
        self._configure_ssh()

    def expand(self):
        """
        Expand environment variables in the host configuration.
        """
        if self.user:
            self.user = os.path.expandvars(self.user)

        if self.sshKey:
            self.sshKey = os.path.expandvars(self.sshKey)

    def _configure_ssh(self) -> None:
        """
        Configure the SSH connection to the host. This method makes connection timeouts configurable
        through the SSH config file.
        """
        with open(os.path.expanduser('~/.ssh/config'), mode='a+', encoding='utf-8') as ssh_config_f:
            ssh_config_f.seek(0)

            _conf = ssh_config_f.read().strip()

            if f'Host {self.address}' in _conf:
                log.debug(f'SSH connection to {self.address} already configured')
                return None

            # Go to the end of the file.
            ssh_config_f.seek(
                0,
                os.SEEK_END
            )

            # Now write the new entry to the ~/.ssh/config for this particular host.
            if _conf != '':
                ssh_config_f.write('\n')

            ssh_config_f.write(f'Host {self.address}\n\tConnectTimeout {self.timeout}\n\tStrictHostKeyChecking no\n\tIdentityFile ~/.ssh/{self.name}\n')

        # Write the SSH key to the ~/.ssh directory.
        if self.sshKey is not None:
            with open(os.path.expanduser(f'~/.ssh/{self.name}'), mode='w', encoding='utf-8') as ssh_key_f:
                log.debug(f'Writing SSH key to ~/.ssh/{self.name} for host at address {self.address}')
                ssh_key_f.write(self.sshKey + '\n')

            os.chmod(os.path.expanduser(f'~/.ssh/{self.name}'), 0o600)

        log.info(f'Configured SSH connections to host {self.address} with a timeout of {self.timeout} seconds')

    # Minor helper functions to reduce code duplication.
    def to_db_entry(self) -> Dict:
        """
        Convert the host configuration to a flat database record-format.

        Returns:
            Dict: This instance as a database record.
        """
        host_dict = {
            'name': self.name,
            'address': self.address,
            'protocol': self.protocol,
            'port': self.port,
            'hypervisor': self.hypervisor,
            'cpu': self.resources.cpu if self.resources is not None and self.resources.cpu is not None else 0,
            'memory': self.resources.memory if self.resources is not None and self.resources.memory is not None else 0,
            'storage': self.resources.storage if self.resources is not None and self.resources.storage is not None else 0
        }

        return host_dict


@define
class CloudInit:
    """
    Cloud-init configuration options.
    """
    user_data: str
    meta_data: str
    network_data: str
    vendor_data: str


@define
class HostReplacementStrategy:
    """
    Host replacement strategy configuration options.
    """
    strategy: str
    maxUnavailable: int
    maxSurge: int


@define
class Network:
    """
    Network configuration options.
    """
    dhcp: bool     # true, false
    gateway: str  # 192.168.1.1
    netmask: str  # 255.255.255.0
    subnet: str   # 192.168.1.0
    addressRange: str | None = ib(default=None) # e.g., if dhcp is true, '192.168.1.2-192.168.1.59'

    def __attrs_post_init__(self):
        """
        Post-initialization method to expand environment variables.
        """
        if self.dhcp and self.addressRange is None:
            log.error(f'Address range must be provided if DHCP is enabled')
            sys.exit(1)


@define
class ScaleStrategy:
    """
    Scale strategy configuration options.
    """
    min: int
    max: int
    desired: int
    increment: int
    cooldown: int
    method: str
    targetUtilization: Dict[str, int]


@define
class AutoscalingGroup:
    """
    Autoscale group configuration options.
    """
    image: str
    domainName: str
    imageMigrationType: str
    cloudInit: CloudInit
    hosts: List[Host]
    replacement: HostReplacementStrategy
    networking: Network
    scaling: ScaleStrategy


@define(frozen=False)
class AutoscalingGroups:
    """
    Because keys are variable, we need to define a custom init method for autoscaling groups.
    """

    # https://www.attrs.org/en/stable/init.html#custom-init
    def __init__(self, **kwargs: Dict[str, AutoscalingGroup]):
        for key, value in kwargs.items():
            # This ends up being like,
            # asg_1: AutoscalingGroup
            # asg_2: AutoscalingGroup
            # etc. But we don't know how many ASGs users will configure so we can't statically type the keys.
            setattr(self, key, value)


@define
class Autoscale:
    """
    Autoscale configuration options.
    """
    hosts: List[Host]
    groups: AutoscalingGroups


@define
class Healthcheck:
    """
    Healthcheck configuration options.
    """
    host: str
    port: int

@define
class Controller:
    """
    Controller configuration options.
    """
    mode: str
    pidFile: str
    databases: Databases
    platform: Platform
    reconciliation: Reconciliation
    autoscale: Autoscale
    healthcheck: Healthcheck


@define
class Config:
    """
    Parse config files of version v1alpha1.
    """
    version: str
    controller: Controller

    @classmethod
    def from_dict(cls, config: dict) -> Config:
        """
        Create a Config object from a dictionary.

        Args:
            config (dict): The config dictionary.

        Returns:
            Config: The Config object.
        """
        return structure(
            config,
            cls
        )


"""
Interface with the Kubernetes autoscaler.
"""


from __future__ import annotations

import logging

from typing import TYPE_CHECKING
from setproctitle import setproctitle


if TYPE_CHECKING:
    from typing import Any
    from premiscale.config.v1alpha1 import Config


log = logging.getLogger(__name__)


class KubernetesAutoscaler:
    def __init__(self, config: Config) -> None:
        pass

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        log.debug('Starting Kubernetes autoscaler interface subprocess')
        setproctitle('kubernetes-autoscaler-interface')


"""
Methods for interacting with the in-memory database.
"""


from __future__ import annotations

import logging
import sqlite3

from typing import TYPE_CHECKING
from wrapt import synchronized
from premiscale.metrics.state._base import State


if TYPE_CHECKING:
    from typing import List, Tuple


log = logging.getLogger(__name__)


class Local(State):
    """
    Implement a high-level interface to an in-memory state database.
    """

    def __init__(self) -> None:
        self._connection: sqlite3.Connection
        self._cursor: sqlite3.Cursor

    def is_connected(self) -> bool:
        """
        Check if the connection to the MySQL database is open.

        Returns:
            bool: True if the connection is open.
        """
        return self._connection is not None

    @synchronized
    def run(self, query: str, parameters: Tuple | None = None) -> List:
        """
        Run the in-memory database.
        """
        if parameters:
            q = self._cursor.execute(query, parameters).fetchall()
        else:
            q = self._cursor.execute(query).fetchall()

        self.commit()

        return q

    @synchronized
    def open(self, database: str = 'file::memory:?cache=shared') -> None:
        """
        Open a connection to a SQLite database. Defaults to an in-memory database.

        Args:
            database (str): Path to the SQLite database. Defaults to ':memory:'.
        """
        log.debug('Opening connection to in-memory SQLite database')
        self._connection = sqlite3.connect(
            database=database,
            check_same_thread=False
        )
        self._cursor = self._connection.cursor()
        log.debug('Connection to in-memory SQLite database opened')

    @synchronized
    def close(self) -> None:
        """
        Close the connection to the SQLite database.
        """
        log.debug('Closing connection to in-memory SQLite database')
        self._connection.close()

    @synchronized
    def commit(self) -> None:
        """
        Commit any changes to the SQLite database.
        """
        log.debug('Committing changes to in-memory SQLite database')
        self._connection.commit()

    @synchronized
    def rollback(self) -> None:
        """
        Rollback any changes to the SQLite database.
        """
        log.debug('Rolling back changes to in-memory SQLite database')
        self._connection.rollback()

    @synchronized
    def initialize(self) -> None:
        """
        Initialize the SQLite database.
        """
        log.debug('Initializing in-memory SQLite database')
        self._cursor.execute(
            'CREATE TABLE IF NOT EXISTS hosts (name TEXT, address TEXT, protocol TEXT, port INTEGER, hypervisor TEXT, cpu INTEGER, memory INTEGER, storage INTEGER)'
        )
        self._cursor.execute(
            'CREATE TABLE IF NOT EXISTS vms (host TEXT, name TEXT, cores INTEGER, memory INTEGER, storage INTEGER)'
        )
        self._cursor.execute(
            'CREATE TABLE IF NOT EXISTS asgs (name TEXT)'
        )
        self.commit()
        log.debug('In-memory SQLite database initialized')

    ## Hosts

    @synchronized
    def get_host(self, name: str, address: str) -> Tuple | None:
        """
        Get a host record.

        Args:
            name (str): name of host to retrieve.
            address (str): IP address of the host.

        Returns:
            Tuple | None: Host record, if it exists. Otherwise, None.
        """
        entries = self._cursor.execute(
            'SELECT * FROM hosts WHERE name = ? AND address = ?',
            (name, address)
        ).fetchall()

        if len(entries) > 1:
            log.error(f'Expected 1 host record, got {len(entries)}. Returning the first entry.')
            log.debug(f'Host records: {entries}')
        elif len(entries) == 0:
            log.error('No host records found. Returning None.')
            return None

        return entries[0]

    @synchronized
    def host_create(self, name: str, address: str, protocol: str, port: int, hypervisor: str, cpu: int, memory: int, storage: int) -> bool:
        """
        Create a host record.

        Args:
            name (str): name to give host.
            address (str): IP address of the host.
            protocol (str): protocol to use for communication.
            port (int): port to communicate over.
            hypervisor (str): hypervisor to use for VM management.
            cpu (int): number of CPUs available.
            memory (int): amount of memory available.
            storage (int): amount of storage available.

        Returns:
            bool: True if action completed successfully.
        """
        self._cursor.execute(
            'INSERT INTO hosts (name, address, protocol, port, hypervisor, cpu, memory, storage) VALUES (?, ?, ?, ?, ?, ?, ?, ?)',
            (name, address, protocol, port, hypervisor, cpu, memory, storage)
        )
        self.commit()
        return True

    @synchronized
    def host_delete(self, name: str, address: str) -> bool:
        """
        Delete a host record.

        Args:
            name (str): name of host to delete the record for.
            address (str): IP address of the host.

        Returns:
            bool: True if action completed successfully.
        """
        self._cursor.execute(
            'DELETE FROM hosts WHERE name = ? AND address = ?',
            (name, address)
        )
        self.commit()
        return True

    @synchronized
    def host_update(self, name: str, address: str, protocol: str, port: int, hypervisor: str, cpu: int, memory: int, storage: int) -> bool:
        """
        Update a host record.

        Args:
            name (str): name to give host.
            address (str): IP address of the host.
            protocol (str): protocol to use for communication.
            port (int): port to communicate over.
            hypervisor (str): hypervisor to use for VM management.
            cpu (int): number of CPUs available.
            memory (int): amount of memory available.
            storage (int): amount of storage available.

        Returns:
            bool: True if action completed successfully.
        """
        self._cursor.execute(
            'UPDATE hosts SET protocol = ?, port = ?, hypervisor = ?, cpu = ?, memory = ?, storage = ? WHERE name = ? AND address = ?',
            (protocol, port, hypervisor, cpu, memory, storage, name, address)
        )
        self.commit()
        return True

    @synchronized
    def host_exists(self, name: str, address: str) -> bool:
        """
        Check if a host exists in the database.

        Args:
            name (str): name of host to check for.
            address (str): IP address of the host.

        Returns:
            bool: True if the host exists.
        """
        return self._cursor.execute(
            'SELECT * FROM hosts WHERE name = ? AND address = ?',
            (name, address)
        ).fetchone() is not None

    @synchronized
    def host_report(self) -> List:
        """
        Get a report of currently-managed hosts.

        Returns:
            List: List of hosts and the VMs on them.
        """
        return self._cursor.execute(
            'SELECT * FROM hosts'
        ).fetchall()

    ## VMs

    @synchronized
    def vm_create(self, host: str, vm_name: str, cores: int, memory: int, storage: int) -> bool:
        """
        Create a host record.

        Args:
            host (str): host on which to provision the VM.
            vm_name (str): name to give the new VM.
            cores (int): number of cores to allocate.
            memory (int): amount of memory to allocate.
            storage (int): amount of storage to allocate.

        Returns:
            bool: True if action completed successfully.
        """
        self._cursor.execute(
            'INSERT INTO vms (host, name, cores, memory, storage) VALUES (?, ?, ?, ?, ?)',
            (host, vm_name, cores, memory, storage)
        )
        self.commit()
        return True

    @synchronized
    def vm_delete(self, host: str, vm_name: str) -> bool:
        """
        Delete a VM on a specified host.

        Args:
            host (str): host on which to delete the VM.
            vm_name (str): name of VM to delete.

        Returns:
            bool: True if action completed successfully.
        """
        self._cursor.execute(
            'DELETE FROM vms WHERE host = ? AND name = ?',
            (host, vm_name)
        )
        self.commit()
        return True

    @synchronized
    def vm_report(self, host: str | None = None) -> List:
        """
        Get a report of VMs presently-managed on a host.

        Args:
            host (str | None): Name of host on which to retrieve VM entries. If None, return all VMs on all hosts. Defaults to None.

        Returns:
            List: List of VMs on the host, if host was specified; otherwise, all VMs on all hosts.
        """
        if host is None:
            return self._cursor.execute(
                'SELECT * FROM vms'
            ).fetchall()

        return self._cursor.execute(
            'SELECT * FROM vms WHERE host = ?',
            (host,)
        ).fetchall()

    ## ASGs

    @synchronized
    def asg_create(self, name: str) -> bool:
        """
        Create an autoscaling group.

        Args:
            name (str): Name to give the ASG.

        Returns:
            bool: True if action completed successfully.
        """
        self._cursor.execute(
            'INSERT INTO asgs (name) VALUES (?)',
            (name,)
        )
        self.commit()
        return True

    @synchronized
    def asg_delete(self, name: str) -> bool:
        """
        Delete an autoscaling group.

        Args:
            name (str): Name of ASG to delete.

        Returns:
            bool: True if action completed successfully.
        """
        self._cursor.execute(
            'DELETE FROM asgs WHERE name = ?',
            (name,)
        )
        self.commit()
        return True

    @synchronized
    def asg_add_vm(self, host: str, vm_name: str) -> bool:
        """
        Add a VM on a host to an autoscaling group.

        Args:
            host (str): Name of host on which the VM resides.
            vm_name (str): Name of VM to add to ASG.

        Returns:
            bool: True if action completed successfully.
        """
        self._cursor.execute(
            'INSERT INTO asgs (name) VALUES (?)',
            (vm_name,)
        )
        self.commit()
        return True

    @synchronized
    def asg_remove_vm(self, host: str, vm_name: str) -> bool:
        """
        Remove a VM on a host from an ASG.

        Args:
            host (str): Name of host on which the VM resides.
            vm_name (str): Name of VM to remove from ASG.

        Returns:
            bool: True if action completed successfully.
        """
        self._cursor.execute(
            'DELETE FROM asgs WHERE name = ?',
            (vm_name,)
        )
        self.commit()
        return True

    @synchronized
    def get_asg_vms(self, asg_name: str, host: str | None = None) -> List:
        """
        Get all VMs in an autoscaling group, optionally filtering by host.

        Args:
            asg_name (str): Name of ASG to retrieve VMs from.
            host (str | None): Optionally specify the name of host by which to filter the autoscaling group VMs by. Defaults to None.

        Returns:
            List: List of VMs in the ASG.
        """
        if host is None:
            return self._cursor.execute(
                'SELECT * FROM vms WHERE asg_name = ?',
                (asg_name,)
            ).fetchall()

        return self._cursor.execute(
            'SELECT * FROM vms WHERE name = ? AND host = ?',
            (asg_name, host)
        ).fetchall()

    @synchronized
    def asg_report(self, vm_enabled: bool = False) -> List:
        """
        Get a report of current autoscaling groups' standings. Optionally enable VMs be returned on hosts as well.

        Args:
            vm_enabled (bool, optional): Return VMs on hosts as well. Defaults to False as it's a more expensive operation.

        Returns:
            List: List of autoscaling groups and their VMs, if enabled
        """
        if vm_enabled:
            return self._cursor.execute(
                'SELECT * FROM asgs'
            ).fetchall()

        return self._cursor.execute(
            'SELECT * FROM asgs'
        ).fetchall()


"""
Implement a Libvirt connection to a Qemu-based hypervisor/host.
"""


from __future__ import annotations

import logging
import os

from attrs import define
from attr import ib
from cattr import structure
from typing import TYPE_CHECKING, List
from datetime import datetime, timezone

if TYPE_CHECKING:
    from typing import Dict, Tuple


log = logging.getLogger(__name__)


# Schemas for parsing retrieved hypervisor objects.

@define
class vCPU:
    """
    A dataclass for storing CPU statistics.
    """
    state: int
    time: int
    wait: int
    delay: int


@define
class Net:
    """
    A dataclass for storing network interface statistics.
    """
    name: str
    rx_bytes: int
    rx_pkts: int
    rx_errs: int
    rx_drop: int
    tx_bytes: int
    tx_pkts: int
    tx_errs: int
    tx_drop: int


@define
class Block:
    """
    A dataclass for storing block device statistics.
    """
    name: str
    path: str
    backingIndex: int
    rd_reqs: int
    rd_bytes: int
    rd_times: int
    wr_reqs: int
    wr_bytes: int
    wr_times: int
    fl_reqs: int
    fl_times: int
    allocation: int
    capacity: int
    physical: int


@define
class DomainStats:
    """
    A dataclass for storing domain statistics. This is a normalized version of the stats returned by the libvirt API.

    This class also provides methods for converting the domain statistics into a compatible format for TinyFlux and InfluxDB.
    """

    # Name and host are enough to uniquely identify a domain.
    name: str
    host: str
    address: str

    # State
    state_state: int
    state_reason: int

    # CPU
    cpu_time: int
    cpu_user: int
    cpu_system: int
    cpu_cache_monitor_count: int
    cpu_haltpoll_success_time: int
    cpu_haltpoll_fail_time: int

    balloon_rss: int

    # vCPU
    vcpu_current: int
    vcpu_maximum: int
    # TODO: https://github.com/python-attrs/attrs/issues/1298
    vcpu: List[vCPU]

    # net
    net: List[Net]

    # Block devices
    block: List[Block]
    dirtyrate_calc_status: int
    dirtyrate_calc_start_time: int
    dirtyrate_calc_period: int

    # Ballon fields with defaults (it doesn't seem every hypervisor will have these fields)
    balloon_current: int | None = ib(default=None)
    balloon_maximum: int | None = ib(default=None)
    balloon_swap_in: int | None = ib(default=None)
    balloon_swap_out: int | None = ib(default=None)
    balloon_major_fault: int | None = ib(default=None)
    balloon_minor_fault: int | None = ib(default=None)
    balloon_unused: int | None = ib(default=None)
    balloon_available: int | None = ib(default=None)
    balloon_usable: int | None = ib(default=None)
    balloon_last_update: int | None = ib(default=None)
    balloon_disk_caches: int | None = ib(default=None)
    balloon_hugetlb_pgalloc: int | None = ib(default=None)
    balloon_hugetlb_pgfail: int | None = ib(default=None)

    # Potentially dependent fields
    net_count: int | None = ib(default=None)
    block_count: int | None = ib(default=None)

    # Time of the collection. This is important for time series databases.
    time: datetime = ib(default=datetime.now(timezone.utc))

    def __attrs_post_init__(self) -> None:
        if self.block_count is None:
            self.block_count = len(self.block)

        if self.net_count is None:
            self.net_count = len(self.net)

        if self.vcpu_current != self.vcpu_maximum:
            log.warning(f'vCPU count disparity for {self.name}: {self.vcpu_current} current != {self.vcpu_maximum} max. ')

    def to_tinyflux(self) -> Tuple[Dict, Dict, Dict, Dict]:
        """
        Convert the domain statistics into a compatible format for TinyFlux Point objects.

        In the process, 4 different points are created for the CPU, memory, network, and block device statistics.

        Block devices are a bit more complex than the other stats, so we'll break down the schema for them here.

        Block
        =====

            Total physical allocation is the sum of all the physical block devices' 'block.physical' attributes.
            This metric is grouped by block devices' mount points.

            Example:

                {
                    '/var/lib/libvirt/images_utilization': int,
                    '/var/lib/libvirt/images2_utilization': int
                }

            Later on, this metric can be used to determine the total physical allocation of a given mount point by VMs,
            so scheduling can take this into account when placing VMs.

        Returns:
            Tuple[Dict, Dict, Dict, Dict]: all of the concatenated domain statistics on which we can scale on with some
                additional fields. This object takes the following schema

            (
                {
                    # Name of the measurement. This is the name of the table in the database.
                    'measurement': str,

                    # Time of the measurement. It is important this data is stored alongside the collection time.
                    'time': datetime,

                    # Tag attributes that are searchable in the database
                    'tags': Dict[str, str],

                    # Actual data
                    'fields': Dict[str, int | float]
                },
                ... 4 times
            )
        """

        # Put together a record of the domain statistics that's palatable for TinyFlux.

        _cpu_datum: Dict = {
            'measurement': 'cpu',
            'time': self.time,
            'tags': {
                'name': self.name,
                'host': self.host,
                'state': str(self.state_state),
                'reason': str(self.state_reason)
            },
            'fields': {
                # Roughly speaking, these are the four main categories of stats we're interested in autoscaling virtual machines on.
                # Actual CPU utilization percentage is the difference between two consecutive differences between the total CPU time
                # and the sum of user and system time over some interval.
                # $\max(\frac{1}{I}\left(\frac{\text{cpu_time}_1 - (\text{cpu_user}_1 - \text{cpu_system}_1)}{\text{vcpu_current}_1}-\frac{\text{cpu_time}_2 - (\text{cpu_user}_2 - \text{cpu_system}_2)}{\text{vcpu_current}_2}\right), 0)
                'total_cpu_utilization': self.cpu_time - (self.cpu_user + self.cpu_system),
                'cpu_time': self.cpu_time,
                'cpu_user': self.cpu_user,
                'cpu_system': self.cpu_system,
                'vcpu_current': self.vcpu_current,
                'vcpu_maximum': self.vcpu_maximum,
            }
        }

        _memory_datum: Dict = {
            'measurement': 'memory',
            'time': self.time,
            'tags': {
                'name': self.name,
                'host': self.host,
                'state': str(self.state_state),
                'reason': str(self.state_reason)
            },
            'fields': {
                # Memory utilization is the difference between the current and maximum balloon values.
                'total_memory_utilization': round(self.balloon_current / self.balloon_maximum * 100, 2) if self.balloon_current is not None and self.balloon_maximum is not None else -1
            }
        }

        _net_datum: Dict = {
            'measurement': 'net',
            'time': self.time,
            'tags': {
                'name': self.name,
                'host': self.host,
                'state': str(self.state_state),
                'reason': str(self.state_reason)
            },
            'fields': {
                'net_count': self.net_count,
                # Sum utilization, errors and drops across all network interfaces. We can use this to autoscale on network and
                # set thresholds for network errors and drops to either trigger a scale verb or affect scheduling of workloads.
                'total_net_utilization': sum(net.rx_bytes + net.tx_bytes for net in self.net),
                'total_net_errors': sum(net.rx_errs + net.tx_errs for net in self.net),
                'total_net_drops': sum(net.rx_drop + net.tx_drop for net in self.net)
            }
        }

        # Calculate the utilization of each network interface.
        for net in self.net:
            # To autoscale on network interface utilization, we can use the sum of the rx_bytes and tx_bytes fields.
            # This said, we don't know the % utilization of the physical NICs on the host from this metric. Virtual
            # NICs are likely never going to be bottleneck intra-host, but physical NICs are.
            _net_datum['fields'][f'{net.name}_utilization'] = net.rx_bytes + net.tx_bytes

        _block_datum: Dict = {
            'measurement': 'block',
            'time': self.time,
            'tags': {
                'name': self.name,
                'host': self.host,
                'state': str(self.state_state),
                'reason': str(self.state_reason)
            },
            'fields': {
                'block_count': self.block_count
            }
        }

        # Calculate the capacity utilization of each block device.
        for block in self.block:
            _block_datum['fields'][f'{block.name}_capacity_utilization'] = round(block.allocation / block.capacity * 100, )

        for mountpoint in set(os.path.dirname(block.path) for block in self.block):
            _block_datum['fields'][f'{mountpoint}_utlization'] = sum(block.physical for _block in self.block if os.path.dirname(_block.path) == mountpoint)

        return (
            _cpu_datum,
            _memory_datum,
            _net_datum,
            _block_datum
        )

    def to_influx(self) -> Dict:
        """
        Convert the domain statistics into a compatible format for InfluxDB.

        Returns:
            Dict: all of the collected domain statistics with some additional fields.

        Raises:
            NotImplementedError: This method has not been implemented yet.
        """
        raise NotImplementedError('This method has not been implemented yet.')


@define
class HostStats:
    """
    A dataclass for storing host time series metrics.
    """


"""
This module is responsible for the metrics API. Scrape metrics from the Prometheus server and return them to the user.
"""


from flask import Flask
from flask_cors import CORS


app = Flask(__name__)
CORS(app)


__all__ = ['app']


from premiscale.api.metrics.routes import create_metrics_api


create_metrics_api(app)


"""
Controller utils.
"""


from __future__ import annotations

import sys
import logging
import json
import os

from enum import Enum
from pathlib import Path


log = logging.getLogger(__name__)


class LogLevel(Enum):
    info = logging.INFO
    error = logging.ERROR
    warn = logging.WARNING
    debug = logging.DEBUG

    def __str__(self):
        return self.name

    @classmethod
    def from_string(cls, s: str) -> LogLevel:
        """
        Convert a string to the enum value.

        Args:
            s (str): key to convert to the enum value.

        Returns:
            LogLevel: Log level object.
        """
        try:
            return cls[s.lower()]
        except KeyError:
            log.error('Must specify an accepted log level')
            sys.exit(1)


def validate_port(number: str | int, port_name: str | None = None) -> int:
    """
    Validates port number as a string or int.

    Args:
        number (str | int): the port number as either an int or a str.
        port_name (str | None): the name of the port (to use in error messages).

    Returns:
        int: the port number, if it passes all checks.
    """
    try:
        _number = int(number)
    except ValueError:
        if port_name:
            log.error(f'expected a valid port number "{port_name}", received: "{number}"')
        else:
            log.error(f'expected a valid port number, received: "{number}"')
        sys.exit(1)

    if 0x0 > _number > 0xFFFF:
        if port_name:
            log.error(f'port "{port_name}" must be in range 0 < port < 65535, received: "{number}"')
        else:
            log.error(f'port must be in range 0 < port < 65535, received: "{number}"')
        sys.exit(1)

    return _number


def write_json(data: dict, path: str) -> None:
    """
    Write a dictionary to a JSON file.

    Args:
        data (dict): the dictionary to write.
        path (str): the path to write the file to.
    """
    path_e = os.path.expandvars(path)

    try:
        with open(path_e, 'w', encoding='utf-8') as f:
            log.debug(f'Writing JSON to file: {path_e}')
            json.dump(data, f)
    except (FileNotFoundError, PermissionError) as msg:
        log.error(f'Failed to write JSON file, received: {msg}')
    except OSError as msg:
        log.error(f'Failed to write JSON file {path_e}, received: {msg}. Check your path and permissions')


def read_json(path: str) -> dict | None:
    """
    Read a JSON file and return the data as a dictionary.

    Args:
        path (str): the path to the JSON file.

    Returns:
        dict | None: the data from the JSON file, or None if the file does not exist.
    """
    try:
        if Path(path).exists():
            with open(path, 'r', encoding='utf-8') as f:
                return json.load(f)
        return None
    except (FileNotFoundError, PermissionError) as msg:
        log.error(f'Failed to read JSON file, received: {msg}')
        return None


"""
Parse a configuration file, or create a default one.
"""


import yaml
import logging
import sys
import json

from pathlib import Path
from importlib import resources
from yamale import make_schema, make_data, validate as yamale_validate
from premiscale.config.v1alpha1 import Config
from premiscale.config import build_config_from_version


log = logging.getLogger(__name__)


__all__ = [
    'configParse',
    'validateConfig'
]


def configParse(configPath: str) -> Config:
    """
    Parse a config file and return it as a Config-object. If the file does not exist, create a default one.

    Args:
        configPath (str): path to the config file.

    Returns:
        Config: The parsed config file.
    """

    # Drop a default config for parsing if one was not provided by the user.
    if not Path(configPath).exists():
        makeDefaultConfig(configPath)

    with open(configPath, 'r', encoding='utf-8') as f:
        try:
            _loaded_config = yaml.safe_load(
                f.read().rstrip()
            )

            # Validate the config file against the schema.
            if 'version' not in _loaded_config:
                log.error('Config file is missing a version field')
                sys.exit(1)

            if not validateConfig(configPath, version=_loaded_config['version']):
                sys.exit(1)

            # Parse the config into a Config object, now.
            _config = build_config_from_version(_loaded_config['version']).from_dict(_loaded_config)
        except (yaml.YAMLError, KeyError) as e:
            log.error(f'Error parsing config file: {e}')
            sys.exit(1)

    log.debug(f'Successfully parsed config version {_config.version}: {json.dumps(_loaded_config)}')

    return _config


def validateConfig(configPath: str, version: str = 'v1alpha1', strict: bool = True) -> bool:
    """
    Validate users' config files against our schema.

    Args:
        configPath (str): path to a config file path/name to validate against the schema.
        version (str): the version of the config file to validate. (default: 'v1alpha1')
        strict (bool): whether or not to use strict mode on yamale. (default: True)

    Returns:
        bool: True if the config file is valid.
    """

    if not Path(configPath).exists():
        makeDefaultConfig(configPath)
    else:
        log.debug(f'Found config file at {configPath}')

    try:
        with resources.open_text('premiscale.config.schemas', f'schema.{version}.yaml') as schema_f:
            schema = make_schema(path=schema_f.name)
            data = make_data(path=configPath)

        yamale_validate(
            schema,
            data,
            strict=strict
        )

        log.info(f'Config file at {configPath} is valid')
    except FileNotFoundError as e:
        log.error(f'Could not find file: {e}')
        return False
    except ValueError as e:
        log.error(f'Error validating config file: {e}')
        return False

    return True


def makeDefaultConfig(path: str | Path, default_config: str | Path = 'default.yaml') -> None:
    """
    Make a default config file if one does not exist.

    Args:
        path (str | Path): The default location to create an autoscale configuration file, if it doesn't exist.
        default_config (str | Path): The default config file to use when creating a new config file. (default: 'default.yaml')
    """
    try:
        if not Path(path).parent.exists():
            Path(path).parent.mkdir(parents=True)

        if not Path(path).exists():
            log.debug(f'Config file at {path} does not exist. Creating default config file')

            with resources.open_text('premiscale.config', str(default_config)) as default_config_f, open(str(path), 'x', encoding='utf-8') as f:
                f.write(default_config_f.read().strip())

            log.debug(f'Successfully created default config file at \'{str(path)}\'')
    except PermissionError:
        log.error(f'premiscale does not have permission to install to {str(Path(path).parent)}')
        sys.exit(1)



"""
Serve PremiScale controller healthcheck endpoint when in daemon mode.
"""


from http import HTTPStatus
from flask import Response
from flask import Flask

import logging


log = logging.getLogger(__name__)


def create_healthcheck_api(app: Flask) -> None:
    """
    Create a healthcheck API for a Flask app instance.

    Args:
        app (Flask): Flask app instance.
    """

    @app.get('/healthz')
    def healthcheck() -> Response:
        """
        Quick unauthenticated healthcheck endpoint.

        Returns:
            Response: an object with schema like

            {
                "status": "OK"
            }
        """
        return Response(
            {
                'status': 'OK'
            },
            status=HTTPStatus.OK,
            mimetype='application/json'
        )

    @app.get('/ready')
    def ready() -> Response:
        """
        Quick unauthenticated ready endpoint.

        Returns:
            Response: an object with schema like

            {
                "status": "OK"
            }
        """
        return Response(
            {
                'status': 'OK'
            },
            status=HTTPStatus.OK,
            mimetype='application/json'
        )


"""
Methods for interacting with an in-memory metrics store.
"""


from __future__ import annotations

import logging

from typing import TYPE_CHECKING
from datetime import timedelta, datetime, timezone
from tinyflux import TinyFlux, Point, FieldQuery, TagQuery, TimeQuery
from tinyflux.storages import MemoryStorage, CSVStorage
from premiscale.metrics.timeseries._base import TimeSeries

if TYPE_CHECKING:
    from tinyflux.queries import Query
    from typing import Dict, Tuple


log = logging.getLogger(__name__)


class Local(TimeSeries):
    """
    Implement an interface to storing host metrics in memory.
    """

    def __init__(self, retention: timedelta, file: str | None = None) -> None:
        self.retention: timedelta = retention
        self._connection: TinyFlux
        self.file = file

    def is_connected(self) -> bool:
        """
        Check if the connection to the MySQL database is open.

        Returns:
            bool: True if the connection is open.
        """
        return self._connection is not None

    def open(self) -> None:
        """
        Open a connection to the metrics backend these methods interact with.
        """
        if self.file is not None:
            self._connection = TinyFlux(
                path=self.file,
                storage=CSVStorage
            )
        else:
            self._connection = TinyFlux(
                storage=MemoryStorage
            )

    def close(self) -> None:
        """
        Close the connection to the metrics backend.
        """
        self._connection.close()

    def commit(self) -> None:
        """
        Commit any changes to the database. In this class' case, we do nothing since everything is
        committed by default.
        """
        return None

    def insert(self, data: Dict) -> None:
        """
        Insert a point into the metrics store.

        Args:
            data (Dict): a dictionary containing the data to insert.
        """
        point = Point(**data)

        self._connection.insert(point)
        self._run_retention_policy()

    def insert_batch(self, data: Tuple) -> None:
        """
        Insert a batch of points into the metrics store.

        Args:
            data (Tuple): a tuple of dictionaries containing the data to insert.
        """

        # https://tinyflux.readthedocs.io/en/latest/preparing-data.html
        # This field requires 4 arguments: measurement, time, tags, and fields.
        points = [Point(**datum) for datum in data]

        self._connection.insert_multiple(points)
        self._run_retention_policy()

    def clear(self) -> None:
        """
        Clear the metrics store of all data.
        """
        self._connection.remove_all()

    def _run_retention_policy(self) -> None:
        """
        Run the retention policy on the database, removing points older than the retention policy.
        """
        removed_item_number = 0

        for measurement in ['cpu', 'memory', 'block', 'net']:
            removed_item_number += self._connection.remove(
                TimeQuery() < datetime.now(timezone.utc) - self.retention,
                measurement=measurement
            )

        log.debug(f"Retention removed {removed_item_number} items from the database.")

    def get_all(self, measurement: str | None = None) -> Tuple:
        """
        Get all the data in the metrics store.

        Args:
            measurement (str | None): the measurement to get data for. If None, all data is returned. (Default: None.)

        Returns:
            Tuple: all the data in the metrics store.
        """
        return self._connection.search(
            TimeQuery() > (datetime.now(timezone.utc) - self.retention),
            measurement=measurement,
            sorted=True
        )


"""
ABCs for state storage methods.
"""


from __future__ import annotations

import logging

from typing import TYPE_CHECKING
from abc import ABC, abstractmethod


if TYPE_CHECKING:
    from typing import Any, List, Tuple


log = logging.getLogger(__name__)


class State(ABC):
    """
    An abstract base class with a skeleton interface for state class types.
    """

    @abstractmethod
    def is_connected(self) -> bool:
        """
        Check if the connection to the MySQL database is open.

        Returns:
            bool: True if the connection is open.

        Raises:
            NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def open(self) -> None:
        """
        Open a connection to the state backend these methods interact with.

        Raises:
            NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def close(self) -> None:
        """
        Close the connection to the state backend.

        This method should also dereference any secrets that may be stored in memory.

        Raises:
            NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def commit(self) -> None:
        """
        Commit any changes to the database.

        Raises:
            NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def initialize(self) -> None:
        """
        Initialize the state backend.

        Raises:
            NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError

    def __enter__(self) -> State:
        self.open()
        return self

    def __exit__(self, *args: Any) -> None:
        self.close()
        return

    ## Hosts

    @abstractmethod
    def get_host(self, name: str, address: str) -> Tuple | None:
        """
        Get a host record.

        Args:
            name (str): name of host to retrieve.
            address (str): IP address of the host.

        Returns:
            Tuple | None: Host record, if it exists. Otherwise, None.

        Raises:
            NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def host_create(self, name: str, address: str, protocol: str, port: int, hypervisor: str, cpu: int, memory: int, storage: int) -> bool:
        """
        Create a host record.

        Args:
            name (str): name to give host.
            address (str): IP address of the host.
            protocol (str): protocol to use for communication.
            port (int): port to communicate over.
            hypervisor (str): hypervisor to use for VM management.
            cpu (int): number of CPUs available.
            memory (int): amount of memory available.
            storage (int): amount of storage available.

        Returns:
            bool: True if action completed successfully.

        Raises:
            NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def host_delete(self, name: str, address: str) -> bool:
        """
        Delete a host record.

        Args:
            name (str): name of host to delete the record for.
            address (str): IP address of the host.

        Returns:
            bool: True if action completed successfully.

        Raises:
            NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def host_update(self, name: str, address: str, protocol: str, port: int, hypervisor: str, cpu: int, memory: int, storage: int) -> bool:
        """
        Update a host record.

        Args:
            name (str): name to give host.
            address (str): IP address of the host.
            protocol (str): protocol to use for communication.
            port (int): port to communicate over.
            hypervisor (str): hypervisor to use for VM management.
            cpu (int): number of CPUs available.
            memory (int): amount of memory available.
            storage (int): amount of storage available.

        Returns:
            bool: True if action completed successfully.

        Raises:
            NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def host_exists(self, name: str, address: str) -> bool:
        """
        Check if a host exists in the database.

        Args:
            name (str): name of host to check for.
            address (str): IP address of the host.

        Returns:
            bool: True if the host exists.

        Raises:
            NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def host_report(self) -> List:
        """
        Get a report of currently-managed hosts.

        Returns:
            List: List of hosts and the VMs on them.

        Raises:
            NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError

    ## VMs

    @abstractmethod
    def vm_create(self, host: str, vm_name: str, cores: int, memory: int, storage: int) -> bool:
        """
        Create a host record.

        Args:
            host (str): host on which to provision the VM.
            vm_name (str): name to give the new VM.
            cores (int): number of cores to allocate.
            memory (int): amount of memory to allocate.
            storage (int): amount of storage to allocate.

        Returns:
            bool: True if action completed successfully.

        Raises:
            NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def vm_delete(self, host: str, vm_name: str) -> bool:
        """
        Delete a VM on a specified host.

        Args:
            host (str): host on which to delete the VM.
            vm_name (str): name of VM to delete.

        Returns:
            bool: True if action completed successfully.

        Raises:
            NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def vm_report(self, host: str | None = None) -> List:
        """
        Get a report of VMs presently-managed on a host.

        Args:
            host (str | None): Name of host on which to retrieve VM entries. If None, return all VMs on all hosts. Defaults to None.

        Returns:
            List: List of VMs on the host, or all VMs on all hosts if host is None.

        Raises:
            NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError

    ## ASGs

    @abstractmethod
    def asg_create(self, name: str) -> bool:
        """
        Create an autoscaling group.

        Args:
            name (str): Name to give the ASG.

        Returns:
            bool: True if action completed successfully.

        Raises:
            NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def asg_delete(self, name: str) -> bool:
        """
        Delete an autoscaling group.

        Args:
            name (str): Name of ASG to delete.

        Returns:
            bool: True if action completed successfully.

        Raises:
            NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def asg_add_vm(self, host: str, vm_name: str) -> bool:
        """
        Add a VM on a host to an autoscaling group.

        Args:
            host (str): Name of host on which the VM resides.
            vm_name (str): Name of VM to add to ASG.

        Returns:
            bool: True if action completed successfully.

        Raises:
            NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def asg_remove_vm(self, host: str, vm_name: str) -> bool:
        """
        Remove a VM on a host from an ASG.

        Args:
            host (str): Name of host on which the VM resides.
            vm_name (str): Name of VM to remove from ASG.

        Returns:
            bool: True if action completed successfully.

        Raises:
            NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def get_asg_vms(self, name: str, host: str | None) -> List:
        """
        Get all VMs in an autoscaling group, optionally filtering by host.

        Args:
            name (str): Name of ASG to retrieve VMs from.
            host (str | None): Optionally specify the name of host by which to filter the autoscaling group VMs by.

        Returns:
            List: List of VMs in the ASG.

        Raises:
            NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError

    @abstractmethod
    def asg_report(self, vm_enabled: bool = False) -> List:
        """
        Get a report of current autoscaling groups' standings. Optionally enable VMs be returned on hosts as well.

        Args:
            vm_enabled (bool, optional): Return VMs on hosts as well. Defaults to False as it's a more expensive operation.

        Returns:
            List: List of ASGs.

        Raises:
            NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError


"""
A portion of this module is dedicated to the MetricsCollector class, which oversees visiting every host and collecting
metrics and storing them in the appropriate backend database.

The other portion is dedicated to factory methods for building connections to the time-series and state databases.
"""


from __future__ import annotations

import logging

from typing import TYPE_CHECKING
from concurrent.futures import ThreadPoolExecutor
from setproctitle import setproctitle
from cattrs import unstructure
from time import sleep
from datetime import datetime, timedelta
from premiscale.hypervisor import build_hypervisor_connection


if TYPE_CHECKING:
    from typing import Iterator, List, Tuple
    # TODO: Update this to 'from premiscale.config._config import ConfigVersion as Config' once an ABC for Host is implemented.
    from premiscale.config.v1alpha1 import Config, Host
    from premiscale.metrics.state._base import State
    from premiscale.metrics.timeseries._base import TimeSeries


log = logging.getLogger(__name__)


def build_timeseries_connection(config: Config) -> TimeSeries:
    """
    Build a metrics collection class instance for storing metrics  hosts. Only one of
    these loops is created; metrics are published to influxdb for retrieval and query
    by the ASG loop, which evaluates on a per-ASG basis whether Actions need to be taken.

    Args:
        config (Config): The configuration object.

    Returns:
        TimeSeries: A time-series database interface.

    Raises:
        ValueError: If the time-series database type is unknown.
    """
    match config.controller.databases.timeseries.type:
        case 'memory':
            log.debug(f'Using local memory for time series database')
            from premiscale.metrics.timeseries.local import Local
            from datetime import timedelta

            return Local(
                retention=timedelta(seconds=config.controller.databases.timeseries.trailing),
                file=config.controller.databases.timeseries.dbfile
            )
        case 'influxdb':
            log.debug(f'Using InfluxDB for time series database')
            from premiscale.metrics.timeseries.influxdb import InfluxDB

            connection = unstructure(config.controller.databases.timeseries.connection)
            del(connection['type'])

            return InfluxDB(**connection)
        case _:
            raise ValueError(f'Unknown timeseries database type: {config.controller.databases.timeseries.type}')


def build_state_connection(config: Config) -> State:
    """
    Build a state collection class.

    Args:
        config (Config): The configuration object.

    Returns:
        State: A state database interface.

    Raises:
        ValueError: If the state database type is unknown.
    """
    match config.controller.databases.state.type:
        case 'memory':
            # SQLite
            from premiscale.metrics.state.local import Local

            return Local()
        case 'mysql':
            from premiscale.metrics.state.mysql import MySQL

            connection = unstructure(config.controller.databases.state.connection)
            del(connection['type'])

            return MySQL(**connection)
        case _:
            raise ValueError(f'Unknown state database type: {config.controller.databases.state.type}')


class MetricsCollector:
    """
    Oversee visiting every host and collecting metrics and storing them in the appropriate backend database.

    Args:
        config (Config): The configuration object.
        timeseries_enabled (bool): Whether to enable time-series data collection. Defaults to False.
    """
    def __init__(self, config: Config, timeseries_enabled: bool = False) -> None:
        self._timeseriesConnection: TimeSeries | None = None
        self._stateConnection: State
        self.timeseries_enabled = timeseries_enabled
        self.config = config

    def __call__(self) -> None:
        """
        Start the metrics collection subprocess.
        """
        setproctitle('metrics-collector')
        log.debug('Starting metrics collection subprocess')

        self._stateConnection = build_state_connection(self.config)
        self._stateConnection.open()
        self._stateConnection.initialize()

        # Set up database interfaces.
        if self.timeseries_enabled:
            self._timeseriesConnection = build_timeseries_connection(self.config)
            self._timeseriesConnection.open()

        self._initialize_host()
        self._collectMetrics()

    def _initialize_host(self, host: Host | None = None) -> None:
        """
        Ensure hosts are tracked in the database. This is a one-time operation, currently, and no 'host' should be
        specified.

        Eventually, this code should get copied to the collection loop to ensure that hosts are added to the
        database as they are discovered. For now, we'll just add them all at once.

        Args:
            host (Host | None): The host to initialize in the database (for forward compatibility). Defaults to None.
        """

        if host is not None:
            log.warning(f'Host initialization is not yet implemented for individual hosts. All hosts will be initialized at once.')

        _start_time = datetime.now()

        if host is not None:
            _host_dict = host.to_db_entry()

            if not self._stateConnection.host_exists(host.name, host.address):
                self._stateConnection.host_create(**_host_dict)

            _end_time = datetime.now()

            log.debug(f'Host {host.name} initialized in {_end_time - _start_time}')

            return None

        for _h in self:
            if not self._stateConnection.host_exists(_h.name, _h.address):
                self._stateConnection.host_create(
                    **_h.to_db_entry()
                )

        _end_time = datetime.now()
        _total_time = round((_end_time - _start_time).total_seconds(), 2)

        log.debug(f'All hosts initialized in in-memory SQLite database in {_total_time}s')

    def __iter__(self) -> Iterator:
        """
        Create an iterator that visits every host specified in the configuration file.

        Returns:
            Iterator: An iterator that visits every host.
        """
        return iter(self.config.controller.autoscale.hosts)

    def __len__(self) -> int:
        """
        Return the number of hosts specified in the configuration file.

        Returns:
            int: The number of hosts.
        """
        return len(self.config.controller.autoscale.hosts)

    def __getitem__(self, subscript: int | slice) -> List[Host]:
        """
        Get a host by index or slice.

        Args:
            subscript (int | slice): The index or slice of the hosts to retrieve.

        Returns:
            List[Host]: List of hosts at the specified index or within the range of the slice.
        """
        if isinstance(subscript, slice):
            return self.config.controller.autoscale.hosts[subscript.start:subscript.stop:subscript.step]
        else:
            return [self.config.controller.autoscale.hosts[subscript]]

    def _collectMetrics(self) -> None:
        """
        Collect metrics from all hosts and store them in the appropriate backend database.
        """
        # Paginate through hosts to avoid queuing up a large number of jobs to visit hosts.
        maxpages = max(
            1,
            len(self) // self.config.controller.databases.maxHostConnectionThreads + (
                1 if len(self) % self.config.controller.databases.maxHostConnectionThreads > 0 else 0
            ),
            len(self) // (1 if (self.config.controller.databases.hostConnectionQueueSize is None) else self.config.controller.databases.hostConnectionQueueSize) + (
                1 if len(self) % (len(self) if (self.config.controller.databases.hostConnectionQueueSize is None) else self.config.controller.databases.hostConnectionQueueSize) > 0 else 0
            )
        )

        while True:
            collection_run_start = datetime.now()

            for page in range(0, maxpages):

                # Set a lower and upper bound for a slice of hosts to visit from the whole config.
                lower_bound = page * self.config.controller.databases.maxHostConnectionThreads
                _potential_upper_bound = (1 + page) * self.config.controller.databases.maxHostConnectionThreads
                upper_bound = _potential_upper_bound if _potential_upper_bound < len(self) else len(self)

                if lower_bound == upper_bound:
                    log.debug(f'No hosts to visit. Ensure you list managed hosts in the config file')
                    break

                log.debug(f'Collecting metrics for hosts {lower_bound + 1} to {upper_bound} of total {len(self)}')

                # Reset our collection of threads every paginated iteration over hosts.
                threads = []

                with ThreadPoolExecutor(max_workers=self.config.controller.databases.maxHostConnectionThreads) as executor:
                    for host in self[lower_bound:upper_bound]:
                        threads.append(
                            executor.submit(
                                self._collectHostMetrics,
                                host
                            )
                        )

                    for thread in threads:
                        if thread is not None:
                            thread.result()

            # Wait for the next collection interval, or start right away if collection took longer than that interval.

            collection_run_end = datetime.now()
            collection_run_duration = round((collection_run_end - collection_run_start).total_seconds(), 2)

            log.debug(f'Collection run took {collection_run_duration} seconds')

            if collection_run_duration < timedelta(seconds=self.config.controller.databases.collectionInterval).total_seconds():
                actual_sleep = round(self.config.controller.databases.collectionInterval - (collection_run_end - collection_run_start).total_seconds(), 2)
                log.debug(f'Waiting for {actual_sleep} seconds before revisiting every host')
                sleep(actual_sleep)
            else:
                log.warning(f'Collection took longer than the collection interval of {self.config.controller.databases.collectionInterval} seconds. Starting collection immediately')

    def _collectHostMetrics(self, host: Host) -> None:
        """
        Collect metrics for a single host over a readonly Libvirt connection and store them in the appropriate backend database.

        This method is intended to be called with a host argument from a ThreadPoolExecutor-based thread.

        Args:
            host (Host): The host object to collect metrics from.
        """
        with build_hypervisor_connection(host, readonly=True) as host_connection:
            # Exit early; instantiating the connection to the host failed and has already been logged.
            # We'll try again on the next iteration.
            if host_connection is None:
                return None

            log.debug(f'Connection to host {host.name} succeeded, collecting metrics')

            # Diff current state and recorded state and update the state database. We
            # plit reads and writes here to avoid locking the database for too long.
            if self._stateConnection.get_host(host.name, host.address) != (_host_db_entry := host.to_db_entry()):
                self._stateConnection.host_update(
                    **_host_db_entry,
                )

            if self.timeseries_enabled and self._timeseriesConnection is not None:
                # If time series data collection is enabled, collect and store both host and virtual machine time-series data about their performance.
                vms_metrics_db_entry: List[Tuple] = host_connection.statsToMetricsDB()

                log.debug(f'VM metrics for host "{host.name}": {vms_metrics_db_entry}')

                for vm in vms_metrics_db_entry:
                    # vm :: Tuple[Dict, Dict, Dict, Dict]
                    # each Dict is a different measurement by which we can scale on.
                    log.debug(f'Inserting time series metrics for VM: {vm}')
                    self._timeseriesConnection.insert_batch(vm)

                log.debug(f'Time series metrics currently stored: "{self._timeseriesConnection.get_all()}"')


"""
This module contains methods, schemas and classes for parsing and validating configuration files.
"""


from __future__ import annotations

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from typing import Type


def build_config_from_version(version: str) -> Type:
    """
    Build a Config object from a given version.
    """
    match version:
        case 'v1alpha1':
            from premiscale.config.v1alpha1 import Config as ConfigV1Alpha1

            return ConfigV1Alpha1
        case _:
            raise ValueError(f'Unknown config version: {version}')


"""
Custom exceptions for the premiscale package.
"""


from http import HTTPStatus


class RateLimitedError(Exception):
    """
    Raised when an HTTPStatus.TOO_MANY_REQUESTS code is received.
    """
    def __init__(self, message: str ='', delay: float =30.0) -> None:
        super().__init__(message)
        self.message = message
        self.delay = delay

    def __str__(self):
        return f'RateLimitedError(message="{self.message}", code="{HTTPStatus.TOO_MANY_REQUESTS}", "x-rate-limit-reset={self.delay}")'